import os
from cumulusci.tasks.salesforce import BaseSalesforceApiTask
from cumulusci.tasks.bulkdata import LoadData
from cumulusci.tasks.bulkdata.utils import generate_batches
from cumulusci.utils import temporary_dir
from cumulusci.core.config import TaskConfig
from cumulusci.core.utils import import_global
from cumulusci.core.exceptions import TaskOptionsError


class GenerateAndLoadData(BaseSalesforceApiTask):
    """ Orchestrate creating tempfiles, generating data, loading data, cleaning up tempfiles and batching."""

    task_docs = """
    Orchestrate creating tempfiles, generating data, loading data, cleaning up tempfiles and batching.

    CCI has features for generating data and for loading them into orgs. This task pulls them
    together to give some useful additional features, such as storing the intermediate data in
    a tempfile (the default behavior) and generating the data in batches instead of all at
    once (controlled by the `batch_size` option).

    The simplest possible usage is to specify the number of records you'd like generated, a
    mapping file that defines the schema and a data generation task written in Python to actually
    generate the data.

    Use the `num_records` option to specify how many records to generate.
    Use the `mapping` option to specify a mapping file.
    Use 'data_generation_task' to specify what Python class to use to generate the data.'
    Use 'batch_size' to specify how many records to generate and upload in every batch.

    By default it creates the data in a temporary file and then cleans it up later. Specify database_url if you
    need more control than that. The use of both database_url and batch_size together is not currently supported.

    If your generator class makes heavy use of Faker, you might be interested in this patch
    which frequently speeds Faker up. Adding that code to the bottom of your generator file may
    help accelerate it.

    https://sfdc.co/bwKxDD
    """

    task_options = {
        "num_records": {
            "description": "How many records to generate. Precise calcuation depends on the generator.",
            "required": True,
        },
        "batch_size": {
            "description": "How many records to create and load at a time.",
            "required": False,
        },
        "mapping": {"description": "A mapping YAML file to use", "required": True},
        "data_generation_task": {
            "description": "Fully qualified class path of a task to generate the data. Look at cumulusci.tasks.bulkdata.tests.dummy_data_factory to learn how to write them.",
            "required": True,
        },
        "data_generation_options": {
            "description": "Options to pass to the data generator.",
            "required": False,
        },
        "database_url": {
            "description": "A URL to store the database (defaults to a transient SQLite file)",
            "required": "",
        },
    }

    def _init_options(self, kwargs):
        super()._init_options(kwargs)
        self.mapping_file = os.path.abspath(self.options["mapping"])
        if not os.path.exists(self.mapping_file):
            raise TaskOptionsError(f"{self.mapping_file} cannot be found.")
        self.database_url = self.options.get("database_url")
        self.num_records = int(self.options["num_records"])
        self.batch_size = int(self.options.get("batch_size", self.num_records))
        if self.batch_size <= 0:
            raise TaskOptionsError("Batch size should be greater than zero")
        self.class_path = self.options.get("data_generation_task")
        self.data_generation_task = import_global(self.class_path)

        if self.database_url and self.batch_size != self.num_records:
            raise TaskOptionsError(
                "You may not specify both `database_url` and `batch_size` options."
            )

    def _run_task(self):
        with temporary_dir() as tempdir:
            for current_batch_size, index in generate_batches(
                self.num_records, self.batch_size
            ):
                self._generate_batch(
                    self.database_url,
                    tempdir,
                    self.mapping_file,
                    current_batch_size,
                    index,
                )

    def _datagen(self, subtask_options):
        task_config = TaskConfig({"options": subtask_options})
        data_gen_task = self.data_generation_task(
            self.project_config, task_config, org_config=self.org_config
        )
        data_gen_task()

    def _dataload(self, subtask_options):
        subtask_config = TaskConfig({"options": subtask_options})
        subtask = LoadData(
            project_config=self.project_config,
            task_config=subtask_config,
            org_config=self.org_config,
            flow=self.flow,
            name=self.name,
            stepnum=self.stepnum,
        )
        subtask()

    def _generate_batch(self, database_url, tempdir, mapping_file, batch_size, index):
        if not database_url:
            sqlite_path = os.path.join(tempdir, f"generated_data_{index}.db")
            database_url = f"sqlite:///" + sqlite_path
        subtask_options = {
            **self.options,
            "mapping": mapping_file,
            "database_url": database_url,
            "num_records": batch_size,
            "current_batch_number": index,
        }
        self._datagen(subtask_options)
        self._dataload(subtask_options)



from unittest import mock
import os
import unittest

from cumulusci.core.exceptions import TaskOptionsError
from cumulusci.tasks.salesforce import UpdateAdminProfile
from .util import create_task

ADMIN_PROFILE_BEFORE = """<?xml version='1.0' encoding='utf-8'?>
<Profile xmlns="http://soap.sforce.com/2006/04/metadata">
    <applicationVisibilities>
        <application>npsp__Nonprofit_CRM</application>
        <default>true</default>
        <visible>false</visible>
    </applicationVisibilities>
    <classAccess>
        <apexClass>TestClass</apexClass>
        <enabled>false</enabled>
    </classAccess>
    <fieldPermissions>
        <field>Account.TestField__c</field>
        <editable>false</editable>
        <readable>false</readable>
    </fieldPermissions>
    <pageAccesses>
        <apexPage>TestPage</apexPage>
        <enabled>false</enabled>
    </pageAccesses>
    <recordTypeVisibilities>
        <recordType>Account.Business_Account</recordType>
        <default>true</default>
        <personAccountDefault>true</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <recordTypeVisibilities>
        <recordType>Account.HH_Account</recordType>
        <default>false</default>
        <personAccountDefault>false</personAccountDefault>
        <visible>false</visible>
    </recordTypeVisibilities>
    <tabVisibilities>
        <tab>NPSP_Settings</tab>
        <visibility>Hidden</visibility>
    </tabVisibilities>
</Profile>"""

ADMIN_PROFILE_EXPECTED = """<?xml version='1.0' encoding='utf-8'?>
<Profile xmlns="http://soap.sforce.com/2006/04/metadata">
    <applicationVisibilities>
        <application>npsp__Nonprofit_CRM</application>
        <default>true</default>
        <visible>true</visible>
    </applicationVisibilities>
    <classAccess>
        <apexClass>TestClass</apexClass>
        <enabled>true</enabled>
    </classAccess>
    <fieldPermissions>
        <field>Account.TestField__c</field>
        <editable>true</editable>
        <readable>true</readable>
    </fieldPermissions>
    <pageAccesses>
        <apexPage>TestPage</apexPage>
        <enabled>true</enabled>
    </pageAccesses>
    <recordTypeVisibilities>
        <recordType>Account.Business_Account</recordType>
        <default>false</default>
        <personAccountDefault>false</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <recordTypeVisibilities>
        <recordType>Account.HH_Account</recordType>
        <default>true</default>
        <personAccountDefault>true</personAccountDefault>
        <visible>true</visible>
    </recordTypeVisibilities>
    <tabVisibilities>
        <tab>NPSP_Settings</tab>
        <visibility>DefaultOn</visibility>
    </tabVisibilities>
</Profile>"""


class TestUpdateAdminProfile(unittest.TestCase):
    def test_run_task(self):
        task = create_task(
            UpdateAdminProfile,
            {
                "record_types": [
                    {
                        "record_type": "Account.HH_Account",
                        "default": True,
                        "person_account_default": True,
                    }
                ],
                "namespaced_org": True,
            },
        )

        def _retrieve_unpackaged():
            profiles_path = os.path.join(task.tempdir, "profiles")
            admin_profile_path = os.path.join(profiles_path, "Admin.profile")
            os.mkdir(profiles_path)
            with open(admin_profile_path, "w") as f:
                f.write(ADMIN_PROFILE_BEFORE)

        def _check_result():
            with open(
                os.path.join(task.tempdir, "profiles", "Admin.profile"), "r"
            ) as f:
                result = f.read()
            self.assertMultiLineEqual(ADMIN_PROFILE_EXPECTED, result)

        task._retrieve_unpackaged = _retrieve_unpackaged
        task._deploy_metadata = _check_result
        task()

    def test_run_task__record_type_not_found(self):
        task = create_task(
            UpdateAdminProfile,
            {"record_types": [{"record_type": "DOESNT_EXIST"}], "namespaced_org": True},
        )

        def _retrieve_unpackaged():
            profiles_path = os.path.join(task.tempdir, "profiles")
            admin_profile_path = os.path.join(profiles_path, "Admin.profile")
            os.mkdir(profiles_path)
            with open(admin_profile_path, "w") as f:
                f.write(ADMIN_PROFILE_BEFORE)

        task._retrieve_unpackaged = _retrieve_unpackaged
        with self.assertRaises(TaskOptionsError):
            task()

    @mock.patch("cumulusci.salesforce_api.metadata.ApiRetrieveUnpackaged.__call__")
    def test_retrieve_unpackaged(self, ApiRetrieveUnpackaged):
        task = create_task(UpdateAdminProfile)
        task.tempdir = "/tmp"
        task._retrieve_unpackaged()
        ApiRetrieveUnpackaged.assert_called_once()

    def test_deploy_metadata(self):
        task = create_task(UpdateAdminProfile)
        task.tempdir = "/tmp"
        task._get_api = mock.Mock()
        task._deploy_metadata()
        task._get_api.assert_called_once()



import os

from pytest import fixture
from cumulusci.core.github import get_github_api


@fixture
def gh_api():
    return get_github_api("TestOwner", "TestRepo")


@fixture(scope="class", autouse=True)
def restore_cwd():
    d = os.getcwd()
    try:
        yield
    finally:
        os.chdir(d)



import unittest
from unittest.mock import Mock, patch
from cumulusci.salesforce_api.utils import get_simple_salesforce_connection
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci import __version__


class TestSalesforceApiUtils(unittest.TestCase):
    @patch("simple_salesforce.Salesforce")
    def test_connection(self, mock_sf):
        org_config = Mock()
        proj_config = Mock()
        service_mock = Mock()
        service_mock.client_id = "TEST"
        proj_config.keychain.get_service.return_value = service_mock

        get_simple_salesforce_connection(proj_config, org_config)
        mock_sf.assert_called_once_with(
            instance_url=org_config.instance_url,
            session_id=org_config.access_token,
            version=proj_config.project__package__api_version,
        )

        mock_sf.return_value.headers.setdefault.assert_called_once_with(
            "Sforce-Call-Options", "client={}".format(service_mock.client_id)
        )

    @patch("simple_salesforce.Salesforce")
    def test_connection__explicit_api_version(self, mock_sf):
        org_config = Mock()
        proj_config = Mock()
        service_mock = Mock()
        service_mock.client_id = "TEST"
        proj_config.keychain.get_service.return_value = service_mock

        get_simple_salesforce_connection(proj_config, org_config, api_version="42.0")
        mock_sf.assert_called_once_with(
            instance_url=org_config.instance_url,
            session_id=org_config.access_token,
            version="42.0",
        )

        mock_sf.return_value.headers.setdefault.assert_called_once_with(
            "Sforce-Call-Options", "client={}".format(service_mock.client_id)
        )

    @patch("simple_salesforce.Salesforce")
    def test_connection__no_connected_app(self, mock_sf):
        org_config = Mock()
        proj_config = Mock()
        proj_config.keychain.get_service.side_effect = ServiceNotConfigured

        get_simple_salesforce_connection(proj_config, org_config)

        mock_sf.return_value.headers.setdefault.assert_called_once_with(
            "Sforce-Call-Options",
            "client={}".format("CumulusCI/{}".format(__version__)),
        )



import simple_salesforce

from cumulusci import __version__
from cumulusci.core.exceptions import ServiceNotConfigured
from cumulusci.core.exceptions import ServiceNotValid

CALL_OPTS_HEADER_KEY = "Sforce-Call-Options"


def get_simple_salesforce_connection(project_config, org_config, api_version=None):
    sf = simple_salesforce.Salesforce(
        instance_url=org_config.instance_url,
        session_id=org_config.access_token,
        version=api_version or project_config.project__package__api_version,
    )
    try:
        app = project_config.keychain.get_service("connectedapp")
        client_name = app.client_id
    except (ServiceNotValid, ServiceNotConfigured):
        client_name = "CumulusCI/{}".format(__version__)

    sf.headers.setdefault(CALL_OPTS_HEADER_KEY, "client={}".format(client_name))

    return sf
